import ftd3xx
import ftd3xx._ftd3xx_win32 as _ft
import sys, time, timeit, os
import zipfile
import array, struct
import cProfile

samplerate = 25000000 #Hz
capture_time = 1.0 #seconds

def WritePipe(D3XX, pipe, buffer, size):
 if sys.platform == 'linux2':
  pipe -= 0x02
 return D3XX.writePipe(pipe, buffer, size)

def ReadPipe(D3XX, pipe, size):
 if sys.platform == 'linux2':
  pipe -= 0x82
 result = D3XX.readPipeEx(pipe, size)
 return result

def CancelPipe(D3XX, pipe):
 if sys.platform == 'linux2':
  return D3XX.flushPipe(pipe)
 return D3XX.abortPipe(pipe) 
 
def main():
 clock_divider = (100000000 / samplerate) - 1
 numDevices = ftd3xx.createDeviceInfoList()
 if (numDevices == 0):
  print("ERROR: Please check environment setup! No device is detected.")
  return False
 devList = ftd3xx.getDeviceInfoList()
 D3XX = ftd3xx.create(0, _ft.FT_OPEN_BY_INDEX)
 if (D3XX is None):
  print("ERROR: Please check if another D3XX application is open!")
  return False
 devDesc = D3XX.getDeviceDescriptor()
 bUSB3 = devDesc.bcdUSB >= 0x300

 D3XX.setSuspendTimeout(0)
 D3XX.setPipeTimeout(0x82, 10)  

 sendstring = ""
 for counter in range(128): #Set input MUX. Format LSB - MSB: [destination] [dont's care] [source] [0xff] source and destination from 0x00 to oxff, 
  sendstring += chr(counter) + "\x00" + chr(counter) + "\xff"
 sendstring += "\x01\x00\xff\xfe" #0xff 0xfe: Set trigger mode 0x00 0x00 = None, 0x01 0x00 = Continuous mode, 0x02 0x00 = Pattern trigger (TODO!)
 sendstring += chr((clock_divider >> 16 ) & 0xff)+"\x00\xff\xfd"+chr(clock_divider & 0xff)+chr((clock_divider >> 8) & 0xff)+"\xfe\xfd" #[0xfd/0xfe] 0xff: Set clock divider 
 WritePipe(D3XX, 0x02, sendstring, len(sendstring)) 

 full_counter = 0
 overflow_counter = 0
 counter = 0
 data_counter = 0

 raw_file = open("raw.dat", "wb", 1)

 start = timeit.default_timer()
 for a in range(5): # flush buffer: read 5 dummy frames
  result = ReadPipe(D3XX, 0x82, 65536)
 while timeit.default_timer() < start+capture_time: # Capture for 1 second
  result = ReadPipe(D3XX, 0x82, 65536)
  status = D3XX.getLastError()
  if samplerate < 25000:
   time.sleep(0.001)
  raw_file.write(str(result["bytes"]))
  counter += 1
  full_counter = 0
  data_counter += result["bytesTransferred"]
  if result["bytesTransferred"] == 65536: # A frame is 16 bytes minimum, 65536 bytes maximum. Count number of times that frames are maxed out (Buffer overflow?)
   full_counter += 1
   if full_counter == 3:
    overflow_counter += 1
   
 raw_file.close()

 sendstring = "\x00\x00\xff\xfe" #0xff 0xfe: Set trigger mode 0x00 0x00 = None, 0x01 0x00 = Continuous mode, 0x02 0x00 = Pattern trigger (TODO!)
 WritePipe(D3XX, 0x02, sendstring, len(sendstring)) 
 
 print ""
 print "Captured "+str(counter)+" frames, "+str(overflow_counter)+" overflow frames."
 print "Captured "+str(data_counter)+" input bytes."

 if data_counter==0:
  os.remove("raw.dat")
  CancelPipe(D3XX, 0x82)
  D3XX.close()
  D3XX = 0
  return False
 
 zip_file = zipfile.ZipFile("CoLA - "+time.strftime("%d-%m-%Y %H.%M.%S")+".sr", "w")

 zip_file.writestr("version", "2") # a file "version" contains just "2" to indicate version

 metadata = "[global]\r\nsigrok version=0.5.0\r\n\r\n" # write header files of Sigrok ZIP file
 for b in ["A"]:
  metadata += "[device "+b+"]\r\ncapturefile=logic-"+b+"\r\ntotal probes=64\r\nsamplerate="+str(100000000/(100000000/samplerate))+" Hz\r\n" 
  for a in range(32):
   metadata += "probe"+str(a+1)+"=A"+str(a)+"\r\n"
  for a in range(32):
   metadata += "probe"+str(a+33)+"=B"+str(a)+"\r\n"
 metadata += "unitsize=9\r\n\r\n"

 zip_file.writestr("metadata", metadata) #metadata file containing stream information. Unit file of 8 (bytes) results in strange behaviour, so unitsize = 9 is chosen. 1 byte is wasted each sample.

 filecount = 0
 analyzer_state = ["\x00","\x00","\x00","\x00","\x00","\x00","\x00","\x00","\x00","\x00","\x00","\x00"] # current state of all the logic analyzer's 96 inputs
 buffer_1 = [0]*1024*1024*9 # max file is 9MB, and is predeclared for speed.
 bytecounter = 0
  
 raw_file = open("raw.dat", "rb")

 #cp = cProfile.Profile()
 #cp.enable()

 start = timeit.default_timer()

 contents = raw_file.read()
 contents = map("".join, zip(*[iter(contents)]*4)) # convert raw data file into 4 byte chunks

 sys.stdout.write("Processing.")

 for chunk in contents: # assign data according to its identification byte. Optimized for SPEED!
  ident = ord(chunk[3])
  if ident == 0x82:
   analyzer_state[11] = chunk[2]
   analyzer_state[10] = chunk[1]
   analyzer_state[9] = chunk[0]
  elif ident == 0x81:
   analyzer_state[8] = chunk[2]
   analyzer_state[7] = chunk[1]
   analyzer_state[6] = chunk[0]
  elif ident == 0x80:
   analyzer_state[5] = chunk[2]
   analyzer_state[4] = chunk[1]
   analyzer_state[3] = chunk[0]
  elif ident < 0x80:
   analyzer_state[2] = chunk[2]
   analyzer_state[1] = chunk[1]
   analyzer_state[0] = chunk[0]
   for repeat in range(ident+1): # a frame identification lower than 0x80 indicates bytes 23 - 0, and [RLE repeat amount+1]
    buffer_1[bytecounter] = analyzer_state[0]
    buffer_1[bytecounter+1] = analyzer_state[1]
    buffer_1[bytecounter+2] = analyzer_state[2]
    buffer_1[bytecounter+3] = analyzer_state[3]
    buffer_1[bytecounter+4] = analyzer_state[4]
    buffer_1[bytecounter+5] = analyzer_state[5]
    buffer_1[bytecounter+6] = analyzer_state[6]
    buffer_1[bytecounter+7] = analyzer_state[7]
    buffer_1[bytecounter+8] = analyzer_state[8]
    bytecounter += 9
    if bytecounter == 1024*1024*9: # file is full, start a new one!
     buffer_1 = "".join(buffer_1)
     zip_file.writestr("logic-A-"+str(filecount+1), buffer_1, zipfile.ZIP_DEFLATED)
     sys.stdout.write(".")
     filecount += 1
     bytecounter = 0
     buffer_1 = [0]*1024*1024*9
     
 buffer_1 = "".join(buffer_1[:bytecounter]) # write remaining contents to the last file.
 zip_file.writestr("logic-A-"+str(filecount+1), buffer_1, zipfile.ZIP_DEFLATED)

 #cp.disable()
 #cp.print_stats()

 zip_file.close()
 raw_file.close()
 #os.remove("raw.dat")

 print ""
 print "Output " + str((1024*1024*9*filecount)+bytecounter) + " bytes (Decompression ratio: "+ "{:0.1f}".format(((1024*1024*9.0*filecount)+bytecounter)/data_counter) +") in " + "{:0.1f}".format(timeit.default_timer()-start) + " seconds."

 D3XX.close()
 D3XX = 0

if __name__ == "__main__":
 main()
